from __future__ import annotations
import json
from dataclasses import asdict
from typing import List

from aiohttp import web
from aiohttp.web_response import Response

from . import judger, utils, constants, entity


class Controller:
    _judger: judger.Judger

    def __init__(self, target: judger.Judger):
        self._judger = target

    # /info
    async def info(self, request) -> Response:
        res = []
        for i in self._judger.robots.values():
            res.append(entity.RobotDTO(i.status, i.username, i.uuid))
        return utils.response_ok([asdict(item) for item in res])

    # /submit
    async def submit(self, request) -> Response:
        data = await request.post()
        checkpoints: List[entity.CheckpointToProblem] = [entity.CheckpointToProblem(**item) for item in json.loads(data["checkpoints"])]
        rid: int = data["rid"]
        code: str = data["code"]
        res: List[entity.Trace] = await self._judger.submit(checkpoints, code, rid)
        return utils.response_ok([asdict(item) for item in res])

    # /robot/create
    async def robot_create(self, request) -> Response:
        robot = await judger.create_robot()
        await self._judger.register_robot(robot)
        uuid: str = robot.uuid
        return utils.response_ok(uuid)

    # /robot/verify
    async def robot_verify(self, request) -> Response:
        uuid: str = request.query["uuid"]
        if uuid not in self._judger.robots:
            raise entity.COJException("robot不存在")
        res: str = await self._judger.event.robot_verify(self._judger.robots[uuid])
        return utils.response_ok(res)

    # /robot/login
    async def robot_login(self, request) -> Response:
        uuid: str = request.query["uuid"]
        if uuid not in self._judger.robots:
            raise entity.COJException("目标robot不存在")
        robot: entity.Robot = self._judger.robots[uuid]
        robot.username = request.query["username"]
        robot.password = request.query["password"]
        if "code" in request.query:
            code: str = request.query["code"]
            await self._judger.event.robot_login(robot, code)
        else:
            flag = await self._judger.event.robot_login(robot)
            if not flag:
                return utils.response_code(constants.ResponseCode.NEED_VERIFY)
        return utils.response_ok()

    # /link
    async def link(self, request) -> Response:
        option: int = request.query["option"]
        pid: str = request.query["pid"]
        if option == 0:
            await self._judger.data_source.link_all(pid)
        if option == 1:
            robot: str = request.query["robot"]
            if self._judger.data_source.link_exist(pid, robot):
                raise entity.COJException("已经设置目标题目能被该robot揽收")
            await self._judger.data_source.link_add(pid, robot)
        if option == 2:
            robot: str = request.query["robot"]
            if not self._judger.data_source.link_exist(pid, robot):
                raise entity.COJException("未设置目标题目能被该robot揽收")
            await self._judger.data_source.link_del(pid, robot)
        if option == 3:
            data = await self._judger.data_source.link_list(pid)
            return utils.response_ok(data)
        return utils.response_ok()

    # /robot/delete
    async def robot_delete(self, request) -> Response:
        uuid = request.query["uuid"]
        if uuid not in self._judger.robots:
            raise entity.COJException("Robot不存在")
        await self._judger.delete_robot(self._judger.robots[uuid])
        return utils.response_ok()


class HttpServer:
    _judger: judger.Judger
    _controller: Controller

    def __init__(self, target: judger.Judger):
        self._judger = target
        self._controller = Controller(target)

    @web.middleware
    async def _error_middleware(self, request, handler):
        try:
            response = await handler(request)
            if response.status == 404:
                return utils.response_code(constants.ResponseCode.SPACE)
            return response
        except entity.COJException as e:
            return utils.response_message(str(e))
        except Exception as e:
            self._judger.logger.error(utils.get_exception_details(e, self._judger))
            return utils.response_message(utils.get_exception_details(e, self._judger))

    async def start(self):
        app = web.Application(middlewares=[self._error_middleware])
        app.router.add_routes([
            web.get('/info', self._controller.info),
            web.post('/submit', self._controller.submit),
            web.get('/robot/create', self._controller.robot_create),
            web.get('/robot/verify', self._controller.robot_verify),
            web.get('/robot/login', self._controller.robot_login),
            web.get('/robot/delete', self._controller.robot_delete),
            web.get('/link', self._controller.link),
        ])
        runner = web.AppRunner(app)
        await runner.setup()
        site = web.TCPSite(runner, self._judger.host, self._judger.port)
        await site.start()



from __future__ import annotations
import asyncio
from dataclasses import dataclass
from typing import Optional, List, Dict
import aiohttp

from . import constants, utils, judger


@dataclass
class Robot:
    """Robot实体"""

    uuid: str = ""
    """标识符"""

    password: str = ""
    """robot密码"""

    username: str = ""
    """robot账号"""

    status: constants.RobotStatus = constants.RobotStatus.EMPTY
    """robot状态"""

    session: Optional[aiohttp.ClientSession] = None
    """robot的session"""

    queue: Optional[asyncio.Queue] = None
    """评测队列"""


# 抛出的信息将会返回给coj服务端
class COJException(Exception):
    pass


@dataclass
class RobotDTO:
    """RobotDTO"""

    status: int
    """当前状态"""

    username: str
    """robot账号"""

    uuid: str
    """唯一标识符"""


@dataclass
class Trace:
    """Trace"""

    ids: List[int]
    """涉及到的测试点编号"""

    status: constants.TraceStatus
    """状态"""

    username: Optional[str] = None
    """揽收的robot的用户名，status为0时可用"""

    uuid: Optional[str] = None
    """揽收的robot的uuid，status为0时可用"""


@dataclass
class CheckpointToProblem:
    """CheckpointToProblem"""

    extra: str
    """提交给judger时附带信息，一般包含编译器参数等信息"""

    id: int
    """测试点序号"""

    jid: str
    """该测试点所用到的judger的id"""

    memLimit: int
    """内存限制，单位kb，注意，内存和时间限制需要与远程题目对应的一致，否则无法达到效果"""

    nth: int
    """远程题目测试点编号"""

    score: int
    """该测试点分数"""

    target: str
    """远程题目特征"""

    timeLimit: int
    """时间限制，单位ms"""


class CheckpointsPackage:
    rid: int
    """对应的rid"""

    code: str
    """评测代码"""

    checkpoints: Dict[int, CheckpointToProblem]
    """包含的测试点，key为测试点的本地编号"""

    still: List[int]
    """远程测试点中仍然没有完成的抓取的测试点编号"""

    _map: Dict[int, int]

    _judger: judger.Judger

    _robot: Robot

    def __init__(self, rid: int, code: str, checkpoints: List[CheckpointToProblem], _judger: judger.Judger, _robot: Robot):
        self.rid = rid
        self.checkpoints = {}
        self.code = code
        self._map = {}
        self.still = []
        for i in checkpoints:
            self.still.append(i.nth)
            self.checkpoints[i.id] = i
            self._map[i.nth] = i.id
        self._robot = _robot
        self._judger = _judger

    async def reject(
            self,
            status: constants.CheckpointStatus,
            message: str = None,
            log: str = None,
            color: constants.LogColor = constants.LogColor.RED
    ):
        """
        将剩余测试点的状态置为某个状态，并添加日志，这个通常位于评测过程中judger发生了内部错误
        :param status: 要设置的测试点状态
        :param message: 要设置的测试点信息，可选
        :param log: 要附带的log，可选
        :param color: log的颜色，可选，默认为红色
        """
        if log is not None:
            log = f"[{self._judger.jid} - {self._robot.username}] 评测测试点时发生错误：\n{log}\n如下测试点受到影响：{utils.array_to_text(self.still)}"
            self._judger.logger.info(log)
            await self.log(log, color)
        await self.update(self.still, status, message=message)

    async def accept(self):
        await self.update(self.still, constants.CheckpointStatus.JUDGING)
        self._judger.logger.info(f"[{self._judger.jid} - {self._robot.username}] 开始评测：{self.rid}")

    async def log(self, message: str, color: constants.LogColor = constants.LogColor.RED):
        """
        给评测记录发送log
        :param message: log内容
        :param color: log颜色，默认为红色
        """
        await self._judger.http_client.post("/log", data={"rid": self.rid,
                                                          "message": f'<div class="mdui-text-color-{color}">{message}</div>'})

    async def update(
            self,
            ids: List[int],
            status: constants.CheckpointStatus = None,
            message: str = None,
            score: int = None,
            runTime: int = None,
            runMem: int = None,
            done: bool = False
    ):
        """
        更新测试点信息
        :param ids: 要更新的测试点编号
        :param status: 设置测试点为某状态，可选，不填就不设置
        :param message: 设置测试点的信息，可选，不填就不设置
        :param score: 设置测试点的分数，可选，不填就不设置
        :param runTime: 设置测试点的运行时间，单位ms，可选，不填就不设置
        :param runMem: 设置测试点的运行内存，单位mb，可选，不填就不设置
        :param done: 是否在更新完之后从still中删除该测试点
        """
        data = {"rid": self.rid, "ids": utils.array_to_text(ids)}
        if status is not None:
            data["status"] = status
        if message is not None:
            data["message"] = message
        if score is not None:
            data["score"] = score
        if runTime is not None:
            data["runTime"] = runTime
        if runMem is not None:
            data["runMem"] = runMem
        await self._judger.http_client.post("/update", data=data)
        if done:
            for i in ids:
                self.still.remove(self.checkpoints[i].nth)

    def r2l(self, remote_id: int) -> int:
        """
        远程测试点编号转本地测试点编号
        :param remote_id: 远程测试点编号
        :return: 本地测试点编号
        """
        return self._map[remote_id]

    def l2r(self, local_id: int) -> int:
        """
        本地测试点编号转远程测试点编号
        :param local_id: 本地测试点编号
        :return: 远程测试点编号
        """
        return self.checkpoints[local_id].nth



import coj_judger.constants
import coj_judger.entity
import coj_judger.interface
import coj_judger.judger

Judger = coj_judger.judger.Judger
create_robot = coj_judger.judger.create_robot
RobotStatus = coj_judger.constants.RobotStatus
CheckpointStatus = coj_judger.constants.CheckpointStatus
LogColor = coj_judger.constants.LogColor
Robot = coj_judger.entity.Robot
COJException = coj_judger.entity.COJException
CheckpointToProblem = coj_judger.entity.CheckpointToProblem
CheckpointsPackage = coj_judger.entity.CheckpointsPackage
DataInterface = coj_judger.interface.DataInterface
DefaultDataSource = coj_judger.interface.DefaultDataSource
EventInterface = coj_judger.interface.EventInterface



from __future__ import annotations
from typing import Dict
import aiohttp

from . import utils, judger


class HttpClient:
    _session: aiohttp.ClientSession
    _target: judger.Judger

    def __init__(self):
        pass

    async def get(self, url: str, params: Dict) -> None:
        try:
            async with self._session.get(f"{self._target.server}{url}", params=params) as r:
                if r.status == 200:
                    data = await r.json()
                    if data["code"] == 200:
                        return
                    else:
                        self._target.logger.error(
                            f"请求COJ服务端时发生错误：响应代码异常\nGET {self._target.server}{url}\n{data}"
                        )
                else:
                    self._target.logger.error(
                        f"请求COJ服务端时发生错误：HTTP响应代码异常\nGET {self._target.server}{url}\nHTTP响应代码：{r.status}"
                    )
        except Exception as e:
            self._target.logger.error(
                f"请求COJ服务端时发生错误：请求异常\nGET {self._target.server}{url}\n{utils.get_exception_details(e, self._target)}"
            )

    async def post(self, url: str, data: Dict) -> None:
        try:
            async with self._session.post(f"{self._target.server}{url}", data=data) as r:
                if r.status == 200:
                    data = await r.json()
                    if data["code"] == 200:
                        return
                    else:
                        self._target.logger.error(f"请求COJ服务端时发生错误：响应代码异常\nPOST {self._target.server}{url}\n{data}")
                else:
                    self._target.logger.error(
                        f"请求COJ服务端时发生错误：HTTP响应代码异常\nPOST {self._target.server}{url}\nHTTP响应代码：{r.status}"
                    )
        except Exception as e:
            self._target.logger.error(
                f"请求COJ服务端时发生错误：请求异常\nPOST {self._target.server}{url}\n{utils.get_exception_details(e, self._target)}"
            )

    async def register(self, target: judger.Judger) -> bool:
        self._target = target
        self._session = aiohttp.ClientSession(headers={"Authorization": target.key})
        async with self._session.post(
                "{server}/register".format(server=target.server),
                data={"jid": target.jid, "remote": target.remote, "name": target.name}
        ) as r:
            if r.status == 200:
                data = await r.json()
                if data["code"] == 200:
                    return True
                target.logger.error(data)
            return False
